@@ -7,7 +7,7 @@ use strict;
 
 package XML::Compile::Translate;
 use vars '$VERSION';
-$VERSION = '0.90';
+$VERSION = '0.91';
 
 
 # Errors are either in _class 'usage': called with request
@@ -55,8 +55,8 @@ sub new($@)
 sub init($)
 {   my ($self, $args) = @_;
 
-    $self->{nss} = $args->{nss}
-        or panic "no namespace tables";
+    $self->{nss} = $args->{nss} or panic "no namespace tables";
+    $self->{prefixes} ||= {};
 
     $self;
 }
@@ -120,17 +120,28 @@ sub assertType($$$$)
     return if $checker->($value);
 
     error __x"field {field} contains '{value}' which is not a valid {type} at {where}"
-        , field => $field, value => $value, type => $type, where => $where
-        , _class => 'usage';
+      , field => $field, value => $value, type => $type, where => $where
+      , _class => 'usage';
 
 }
 
 sub extendAttrs($@)
-{   my ($self, $in, %add) = @_;
+{   my ($self, $in, $add) = @_;
 
-    # new attrs overrule
-    unshift @{$in->{attrs}},     @{$add{attrs}}     if $add{attrs};
-    unshift @{$in->{attrs_any}}, @{$add{attrs_any}} if $add{attrs_any};
+    if(my $a = $add->{attrs})
+    {   # new attrs overrule old definitions, remove doubles!
+        my (@attrs, %in);
+        my @all = (@{$add->{attrs} || []}, @{$in->{attrs} || []});
+        while(@all)
+        {   my ($type, $code) = (shift @all, shift @all);
+            $in{$type}++ and next;
+            push @attrs, $type => $code;
+        }
+        $in->{attrs} = \@attrs;
+    }
+
+    # don't know, probably not correct: only top-level any?
+    unshift @{$in->{attrs_any}}, @{$add->{attrs_any}} if $add->{attrs_any};
     $in;
 }
 
@@ -612,7 +623,7 @@ sub particle($)
     my $max   = $node->getAttribute('maxOccurs');
 
     unless(defined $min)
-    {   $min = $self->actsAs('WRITER')
+    {   $min = ($self->actsAs('WRITER') || $self->{default_values} ne 'EXTEND')
             && ($node->getAttribute('default') || $node->getAttribute('fixed'))
              ? 0 : 1;
     }
@@ -641,8 +652,12 @@ sub particle($)
     defined $label
         or return ();
 
-    return $self->makeBlockHandler($where, $label, $min, $max, $process,$local)
-        if ref $process eq 'BLOCK';
+    if(ref $process eq 'BLOCK')
+    {   my $key   = $self->keyRewrite($label);
+        my $multi = $self->blockLabel($local, $key);
+        return $self->makeBlockHandler($where, $label, $min, $max
+           , $process, $local, $multi);
+    }
 
     my $required;
     $required = $self->makeRequired($where, $label, $process) if $min!=0;
@@ -654,6 +669,26 @@ sub particle($)
     ( ($self->actsAs('READER') ? $label : $key) => $do);
 }
 
+# blockLabel KIND, LABEL
+# Particle blocks, like `sequence' and `choice', which have a maxOccurs
+# (maximum occurrence) which is 2 of more, are represented by an ARRAY
+# of HASHs.  The label with such a block is derived from its first element.
+# This function determines how.
+#  seq_address       sequence get seq_ prepended
+#  cho_gender        choices get cho_ before them
+#  all_money         an all block can also be repreated in spec >1.1
+#  gr_people         group refers to a block of above type, but
+#                       that type is not reflected in the name
+
+my %block_abbrev = qw/sequence seq_  choice cho_  all all_  group gr_/;
+sub blockLabel($$)
+{   my ($self, $kind, $label) = @_;
+    return $label if $kind eq 'element';
+
+    $label =~ s/^(?:seq|cho|all|gr)_//;
+    $block_abbrev{$kind} . (unpack_type $label)[1];
+}
+
 sub particleGroup($)
 {   my ($self, $tree) = @_;
 
@@ -669,11 +704,14 @@ sub particleGroup($)
 
     my $typename = $self->rel2abs($where, $node, $ref);
 
-    my $dest    = $self->namespaces->find(group => $typename)
+    my $dest  = $self->namespaces->find(group => $typename)
         or error __x"cannot find group `{name}' at {where}"
              , name => $typename, where => $where, _class => 'schema';
 
-    my $group   = $tree->descend($dest->{node}, $dest->{local});
+    local @$self{ qw/elems_qual attrs_qual tns/ }
+       = $self->nsContext($dest);
+
+    my $group = $tree->descend($dest->{node}, $dest->{local});
     return () if $group->nrChildren==0;
 
     $group->nrChildren==1
@@ -683,7 +721,7 @@ sub particleGroup($)
     my $local = $group->currentLocal;
     $local    =~ m/^(?:all|choice|sequence)$/
         or error __x"illegal group member `{name}' at {where}"
-               , name => $local, where => $where, _class => 'schema';
+             , name => $local, where => $where, _class => 'schema';
 
     $self->particleBlock($group->descend);
 }
@@ -1037,7 +1075,8 @@ sub anyElement($$$)
         if $^W && $node->getAttribute('notQName');
 
     my ($yes, $no) = $self->translateNsLimits($namespace, $not_ns);
-    (any => $self->makeAnyElement($where, $handler, $yes, $no, $process, $min, $max));
+    (any => $self->makeAnyElement($where, $handler, $yes, $no
+              , $process, $min, $max));
 }
 
 sub translateNsLimits($$)
@@ -1056,11 +1095,11 @@ sub translateNsLimits($$)
     foreach my $list ($include, $exclude)
     {   my @list;
         if(defined $list && length $list)
-        {   foreach my $url (split " ", $list)
+        {   foreach my $uri (split " ", $list)
             {   push @list
-                 , $url eq '##targetNamespace' ? $tns
-                 : $url eq '##local'           ? ()
-                 : $url;
+                 , $uri eq '##targetNamespace' ? $tns
+                 : $uri eq '##local'           ? ()
+                 : $uri;
             }
         }
         push @return, @list ? \@list : undef;
@@ -1210,7 +1249,8 @@ sub simpleContentExtension($)
         or error __x"base of simpleContent not simple at {where}"
              , where => $where, _class => 'schema';
  
-    $self->extendAttrs($basetype, $self->attributeList($tree));
+    $self->extendAttrs($basetype, {$self->attributeList($tree)});
+
     $tree->currentChild
         and error __x"elements left at tail at {where}"
               , where => $tree->path, _class => 'schema';
@@ -1255,7 +1295,7 @@ sub simpleContentRestriction($$)
 
     $type->{st} = $self->applySimpleFacets($tree, $st, 0);
 
-    $self->extendAttrs($type, $self->attributeList($tree));
+    $self->extendAttrs($type, {$self->attributeList($tree)});
 
     $tree->currentChild
         and error __x"elements left at tail at {where}"
@@ -1290,6 +1330,7 @@ sub complexContent($$)
     my $type  = {};
     my $where = $tree->path . '#cce';
 
+use Data::Dumper;
     if($base ne 'anyType')
     {   my $typename = $self->rel2abs($where, $node, $base);
         my $typedef  = $self->namespaces->find(complexType => $typename)
@@ -1304,15 +1345,17 @@ sub complexContent($$)
 
     my $own = $self->complexBody($tree, $mixed);
 
-    unshift @{$own->{$_}},    @{$type->{$_} || []}
-        for qw/attrs attrs_any/;
+    $self->extendAttrs($type, $own);
 
-    unshift @{$own->{elems}}, @{$type->{elems} || []}
-        if $name eq 'extension';
+    if($name eq 'extension')
+    {   push @{$type->{elems}}, @{$own->{elems} || []};
+    }
+    else # restriction
+    {   $type->{elems} = $own->{elems};
+    }
 
-    $own->{mixed} ||= $type->{mixed};
-
-    $own;
+    $type->{mixed} ||= $own->{mixed};
+    $type;
 }
 
 sub complexMixed($)
@@ -1325,20 +1368,41 @@ sub complexMixed($)
 #
 
 # print $self->rel2abs($path, $node, '{ns}type')    ->  '{ns}type'
-# print $self->rel2abs($path, $node, 'prefix:type') ->  '{ns(prefix)}type'
+# print $self->rel2abs($path, $node, 'prefix:type') ->  '{ns-of-prefix}type'
 
 sub rel2abs($$$)
 {   my ($self, $where, $node, $type) = @_;
     return $type if substr($type, 0, 1) eq '{';
 
     my ($prefix, $local) = $type =~ m/^(.+?)\:(.*)/ ? ($1, $2) : ('', $type);
-    my $url = $node->lookupNamespaceURI($prefix);
+    my $uri = $node->lookupNamespaceURI($prefix);
+    $self->_registerNSprefix($self->{prefixes}, $prefix, $uri) if $uri;
 
     error __x"No namespace for prefix `{prefix}' in `{type}' at {where}"
       , prefix => $prefix, type => $type, where => $where, _class => 'schema'
-        if length $prefix && !defined $url;
+        if length $prefix && !defined $uri;
 
-     pack_type $url, $local;
+    pack_type $uri, $local;
+}
+
+sub _registerNSprefix($$$)
+{   my ($self, $table, $prefix, $uri) = @_;
+
+    return $table->{$uri}{prefix}
+        if $table->{$uri}; # namespace already has a prefix
+
+    my %prefs = map { ($_->{prefix} => 1) } values %$table;
+    my $take;
+    if(!$prefs{$prefix}) { $take = $prefix }
+    elsif(!$prefs{''})   { $take = '' }
+    else
+    {   # prefix already in use; create a new x\d+ prefix
+        my $count = 0;
+        $count++ while exists $prefs{"x$count"};
+        $take    = 'x'.$count;
+    }
+    $self->{prefixes}{$uri} = {prefix => $take, uri => $uri, used => 0};
+    $take;
 }
 
 sub anyType($)
