@@ -7,14 +7,14 @@ use strict;
 
 package XML::Compile::Util;
 use vars '$VERSION';
-$VERSION = '0.90';
+$VERSION = '0.91';
 
 use base 'Exporter';
 
 my @constants  = qw/XMLNS SCHEMA1999 SCHEMA2000 SCHEMA2001 SCHEMA2001i/;
 our @EXPORT    = qw/pack_type unpack_type/;
 our @EXPORT_OK =
-  ( qw/pack_id unpack_id odd_elements block_label type_of_node/
+  ( qw/pack_id unpack_id odd_elements type_of_node/
   , @constants
   );
 our %EXPORT_TAGS = (constants => \@constants);
@@ -50,16 +50,6 @@ sub odd_elements(@)
 }
 
 
-my %block_abbrev = qw/sequence seq_  choice cho_  all all_  group gr_/;
-sub block_label($$)
-{   my ($kind, $label) = @_;
-    return $label if $kind eq 'element';
-
-    $label =~ s/^(?:seq|cho|all|gr)_//;
-    $block_abbrev{$kind} . (unpack_type $label)[1];
-}
-
-
 sub type_of_node($)
 {   my $node = shift or return ();
     pack_type $node->namespaceURI, $node->localName;
